//A place to park some math functions
Struct LineTraceCalc
{
	static double, double, double GetHitAnglePitchRoll(FLineTraceData t, Actor source){
		double ang, pt, roll;

		switch (t.HitType)
		{		
			case TRACE_HitFloor:
			case TRACE_HitCeiling:
				[ang, pt, roll] = GetHitAnglePitchRollFromPlane(t, source);
				break;
			case TRACE_HitWall:
				[ang, pt, roll] = GetHitAnglePitchRollFromLine(t);
				break;
			default:
				return 180+source.angle, 270-source.pitch, 0;
		}

		return ang, pt, roll;
	}

	static double, double, double GetHitAnglePitchRollFromPlane(FLineTraceData t, Actor source){
		bool is3dfloor = t.Hit3DFloor != null;
		SecPlane p;

		if (is3dfloor)
		{
			p = t.HitType == TRACE_HitFloor ? t.Hit3DFloor.top : t.Hit3DFloor.bottom;
		}
		else if (t.HitType == TRACE_HitFloor)
		{
			p = t.HitSector.floorplane;
		}
		else // if (t.HitType == TRACE_HitCeiling)
		{
			p = t.HitSector.ceilingplane;
		}

		let n = p.Normal;
		if (is3dfloor){n=-1*n;}
		
		double ang = 0, pt = -asin(n.Z);

		if (n.XY.Length() > 0)
		{
			ang = atan2(n.Y, n.X);
		}

		double roll = source.Angle + 180 - ang;
		return ang, pt - 90, pt > 0 ? roll : -roll;
	}

	static double, double, double GetHitAnglePitchRollFromLine(FLineTraceData t)
	{
		let delta = t.HitLine.delta;
		double ang = atan2(delta.Y, delta.X);
		return ang + (t.LineSide == Line.front ? -90 : 90), -90, 0;
	}
}

//An inventory item carried by a player that manages the cursor functions
class MWR_Cursor3dInv : Inventory
{
	default
	{
		MWR_Cursor3dInv.CursorRange 4096;
		MWR_Cursor3dInv.CursorTraceFlags TRF_THRUHITSCAN | TRF_NOSKY | TRF_ALLACTORS;
		MWR_Cursor3dInv.Cursor2DClassName 'MWR_Cursor2d';
		MWR_Cursor3dInv.Cursor3DClassName 'MWR_Cursor3d';
	}
	
	protected MWR_Cursor2d _cursor2d;
	protected MWR_Cursor3d _cursor3d;
	protected bool bCursorEnabled;
	protected bool bCursorRequestEnabled;
	protected bool bCursorHit;
	protected FLineTraceData tCursorHit;
	
	int CursorRange;
	int TraceFlags;
	string Cursor2DClassName;
	string Cursor3DClassName;
	
	property CursorRange : CursorRange;
	property CursorTraceFlags : TraceFlags;
	property Cursor2DClassName : Cursor2DClassName;
	property Cursor3DClassName : Cursor3DClassName;

	override void Tick()
	{
		if (Owner == null || Owner.player == null){
			return;
		}
		
		if (bCursorRequestEnabled)
		{
			if (!bCursorEnabled)
			{
				bCursorEnabled=true;
				ShowCursor();
				DoAfterEnable();
				DoWhileEnabled();
			}
			else
			{
				ShowCursor();
				DoWhileEnabled();
			}
		}
		else
		{
			if (bCursorEnabled)
			{
				DoBeforeDisable();
				bCursorEnabled=false;
				HideCursor();
			}
			else
			{
				HideCursor();
			}
		}
		
		Super.Tick();
	}
	
	virtual void DoAfterEnable(){}

	virtual void DoWhileEnabled(){}

	virtual void DoBeforeDisable(){}
	
	//request the cursor to be enabled. Return tells if the cursor is currently enabled.
	bool EnableCursor()
	{
		bCursorRequestEnabled=true;
		return bCursorEnabled;
	}
	
	//request the cursor to be disabled. Return tells if the cursor is currently enabled.
	bool DisableCursor()
	{
		bCursorRequestEnabled=false;
		return bCursorEnabled;
	}
	
	clearscope bool GetIsEnabled()
	{
		return bCursorEnabled;
	}
	
	private void ShowCursor()
	{
		bCursorHit = Owner.LineTrace(Owner.Angle, CursorRange, Owner.Pitch, TraceFlags, Owner.player.viewheight, data: tCursorHit);

		if (bCursorHit)
		{
			double ang, pt, roll;
			[ang, pt, roll] = LineTraceCalc.GetHitAnglePitchRoll(tCursorHit, Owner);

			//pull the cursor back 1 map unit to keep it "outside" what it hit
			vector3 CursorPos = tCursorHit.HitLocation -1 * tCursorHit.HitDir;
			
			if (!_cursor2d)
			{	//spawn new cursor
				_cursor2d = MWR_cursor2d(Spawn(Cursor2DClassName, CursorPos));
			}
			else
			{	//move existing cursor
				_cursor2d.SetOrigin(CursorPos, true);
			}
			
			
			if (_cursor2d)
			{	//adjust cursor
				_cursor2d.Angle = ang;
				_cursor2d.Pitch = pt;
				_cursor2d.Roll = roll;
			}

			//Use 3d cursor on actor
			actor aTarget=GetCursorTarget();
			if (aTarget)
			{
				vector3 CursorPos = aTarget.pos;
				
				if (!_cursor3d)
				{	//spawn new cursor
					_cursor3d = MWR_cursor3d(Spawn(Cursor3DClassName, CursorPos));
				}
				else
				{	//move existing cursor
					if (_cursor3d.pos != CursorPos) _cursor3d.SetOrigin(CursorPos, true);
				}
				
				if (_cursor3d)
				{	//adjust cursor
					double r = aTarget.Radius;
					double h = aTarget.Height;
					
					if (r!=_cursor3d.radius || h!=_cursor3d.height)
					{
						_cursor3d.A_SetSize(r, h, false);
					}
				}
			}
			else
			{
				if (_cursor3d) _cursor3d.Destroy();
			}

		}
		else HideCursor();
	}

	private void HideCursor()
	{
		if (_cursor2d) _cursor2d.Destroy();
		if (_cursor3d) _cursor3d.Destroy();
	}
	
	//get the actor that the cursor is currently pointed at. If ViewLast then last target is returned if cursor is disabled
	actor GetCursorTarget(bool ViewLast = false)
	{
		if (!(ViewLast || bCursorEnabled)) return null;
		
		if (!bCursorHit || tCursorHit.HitType != TRACE_HitActor) return null;
		return tCursorHit.HitActor;
	}
}

//A flat sprite used as a cursor on planes
class MWR_Cursor2d : Actor
{
	Default
	{
		+NOINTERACTION;
		+FLATSPRITE;
		//Scale 0.75;
	}

	States
	{
		Spawn:
			CS2D A -1 Bright; //Cursor A 4-pointer
			//CRSR B -1 Bright; //Cursor B cross
			stop;
	}
	
}

//A 3d model used as a cursor on actors
class MWR_Cursor3d : actor
{
	Default
	{
		Radius 16;
		Height 32;
		+NOINTERACTION
		+NOBLOCKMAP //I thought NOINTERACTION would set this automatically?
	}
	States
	{
		Spawn:
		Flashing:
			CS3D GYCW 4 bright; //only the brightest colors
			loop;
		Blue:
			CS3D B -1 bright;
			stop;
		Cyan:
			CS3D C -1 bright;
			stop;
		Green:
			CS3D G -1 bright;
			stop;
		Magenta:
			CS3D M -1 bright;
			stop;
		Red:
			CS3D R -1 bright;
			stop;
		White:
			CS3D W -1 bright;
			stop;
		Yellow:
			CS3D Y -1 bright;
			stop;
	}
	
	void SetColorFlashing() {SetState(ResolveState("Flashing"));}
	void SetColorBlue() {SetState(ResolveState("Blue"));}
	void SetColorCyan() {SetState(ResolveState("Cyan"));}
	void SetColorGreen() {SetState(ResolveState("Green"));}
	void SetColorMagenta() {SetState(ResolveState("Magenta"));}
	void SetColorRed() {SetState(ResolveState("Red"));}
	void SetColorWhite() {SetState(ResolveState("White"));}
	void SetColorYellow() {SetState(ResolveState("Yellow"));}
	
	override void tick()
	{
		Scale.X = double(Radius) * 2;
		Scale.Y = double(Height) * Level.pixelstretch;

		super.tick();
	}
}
